import numpy as np
from typing import Tuple
from .logging_config import get_module_logger

logger = get_module_logger(__name__)

# ===========================
# 1. Basic Functions
# ===========================



def farrow_interpolator(
    data: np.ndarray,
    mu: float
) -> np.complex64:
    """
    Farrow interpolator (3rd order polynomial implementation).
    Estimate the sample value at the fractional delay mu in [0, 1).

    Args:
        data: a buffer containing the most recent 4 samples.
        mu: fractional delay in [0, 1).

    Returns:
        interpolated_sample: the estimated sample value at the fractional delay mu.
    """ 
    x0, x1, x2, x3 = data
    
    c0 = x1
    c1 = 0.5 * (x2 - x0)
    c2 = x0 - (2.5 * x1) + (2 * x2) - (0.5 * x3)
    c3 = (0.5 * (x3 - x0)) + (1.5 * (x1 - x2))

    interpolated_sample = c0 + (c1 * mu) + (c2 * mu**2) + (c3 * mu**3)

    return interpolated_sample


def gardner_ted(
    y_mid: np.complex64,
    y_prev: np.complex64,
    y_current: np.complex64
) -> float:
    """
    Gardner timing error detector.
    
    Args:
        y_mid: the sample at the estimated midpoint (transition) between symbols.
        y_prev: the sample at the previous optimal sampling instant .
        y_current: the sample at the current optimal sampling instant.

    Returns:
        error: the estimated timing error value.
    """
    error = (y_mid.real * (y_prev.real - y_current.real) + y_mid.imag * (y_prev.imag - y_current.imag))

    return error


def update_buffer(
    buf: np.ndarray,
    iq_data: np.ndarray,
    i: int
) -> np.ndarray:
    """
    Update the interpolation buffer with a sliding window from the input data.

    Args:
        buf: the 4-sample buffer to be updated.
        iq_data: input IQ signal.
        i: the current index for the sliding window.

    Returns:
        buf: the updated buffer.
    """
    start_idx = i - 3
    end_idx = i + 1
    buf[:] = iq_data[start_idx:end_idx]

    return buf

# ===========================
# 2. Composite Function
# ===========================


def symbol_timing_recovery(
    iq_data: np.ndarray,
    samples_per_symbol: int = 2,
    alpha: float = 0.0005,
    beta: float = None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Perform symbol timing recovery using a Farrow interpolator controlled by a Gardner TED and a second-order PI loop.

    Args:
        iq_data: input IQ signal after RRC.
        samples_per_symbol: number of samples per symbol in the input signal.
        alpha: proportional gain for the PI loop.
        beta: integral gain for the PI loop. If None, it will be set to 2*sqrt(alpha).
    
    Returns:
        time_recovery_samples: decimated IQ signal (1 sample per symbol).
        tau_vect: History of the fractional delay.
        dtau_vect: History of the integral term (related to frequency correction).
        timing_error: History of the timing error detected by the Gardner TED.
    """    
    logger.info(
        "Starting Symbol Timing Recovery (Gardner/PI Loop). SPS: %d.",
        samples_per_symbol
    )
    if beta is None:
        beta = 2 * np.sqrt(alpha) 
    logger.info(
        "Loop parameters: alpha = %f, beta = %f.",
        alpha, beta
    )

    N = len(iq_data)
    time_recovery_samples = np.zeros(N//samples_per_symbol, dtype=np.complex64)
    dtau_vect = np.zeros(N)
    tau_vect = np.zeros(N)
    timing_error = np.zeros(N)
    
    tau = 0.0
    dtau = 0.0
    counter = 1
    err = 0.0
    buf = np.zeros(4, dtype=np.complex64)


    y_mid = 0.0 + 0.0j
    y_prev = 0.0 + 0.0j

    t = 0
    i = 4
    j = 0
    for x in range(4):
        buf[x] = iq_data[x]
    
    while i < N -1:
        if tau > 1.0:
            buf = update_buffer(buf, iq_data, i + 1)
            t = tau - 1
            current_sample = farrow_interpolator(buf, t)
        elif tau < 0.0:
            buf = update_buffer(buf, iq_data, i -1)
            t = tau + 1
            current_sample = farrow_interpolator(buf, t)
        else:
            buf = update_buffer(buf, iq_data, i)
            t = tau
            current_sample = farrow_interpolator(buf, t)

        counter += 1
        if counter >= samples_per_symbol:
            counter -= samples_per_symbol
            y_current = current_sample
            time_recovery_samples[j] = current_sample
            j += 1

            err = gardner_ted(y_mid, y_prev, y_current)
            
            dtau += alpha * err
            tau += beta * err
            
            y_prev = y_current
        
        else:
            y_mid = current_sample
            err = 0.0
        
        tau = tau + dtau / samples_per_symbol

        dtau_vect[i] = dtau
        tau_vect[i] = tau
        timing_error[i] = err
        i += 1
            

    logger.info("Symbol Timing Recovery has completed.")
    
    return time_recovery_samples, tau_vect, dtau_vect, timing_error











import numpy as np
from typing import Optional, List
import os
import matplotlib.pyplot as plt

from .logging_config import get_module_logger, enable_logging
from .ddc import digital_down_converter as ddc
from .cr import carrier_error_recovery as cr
from .rrc import apply_rrc_match_filter as rrc
from .timing_recovery import symbol_timing_recovery
from .costas import costas_phase_recovery as costas
from .decode import differential_decode as decode
from .packet_utils import find_packet_start, bits_to_packets
from .fletcher_ecc_save import validate_packet

from .plotting import plot_constellation, plot_eye_diagram

logger = get_module_logger(__name__)

def orbdemod(
    raw_data: np.ndarray,
    freq_orbcomm: float = 137.46e6,
    fs_in: float = 480e6,
    fs_mid: float = 2.4e6,
    fs_out: float = 9600,
    baud_rate: float = 4800.0,
    log_level: str = "INFO",
    plot: bool = False,
    plot_save_dir: Optional[str] = None,
    output_file: Optional[str] = "orbdemod_packets.txt"  
) -> List[str]:
    """
    ORBCOMM Demodulator Pipeline.

    Args:
        raw_data: input voltage signal.
        freq_orbcomm: center frequency of the ORBCOMM signal (local oscillator frequency).
        fs_in: original sampling frequency of the input signal.
        fs_mid: intermediate sampling frequency after first decimation.
        fs_out: final target sampling frequency after second decimation.
        baud_rate: symbol rate (baud).
        log_level: logging level for the pipeline.
        plot: whether to generate and save plots at each stage.
        plot_save_dir: directory to save plots if plotting is enabled.
        output_file: file to save valid decoded hex packets.

    Returns:
        valid_hex_packets: list of valid decoded hex packets.
    """
    enable_logging(level=log_level)
    logger.info("Starting Pipeline with Log Level: %s, Plotting: %s", log_level, plot)

    if plot and plot_save_dir:
        os.makedirs(plot_save_dir, exist_ok=True)

    iq_data, _ = ddc(raw_data, freq_orbcomm, fs_in, fs_mid, fs_out)
    if plot and plot_save_dir:
        plot_constellation(iq_data, sample_skip=1,title="DDC Constellation", save=True, 
                           save_path=os.path.join(plot_save_dir, 'iq_ddc.png'))
    
    iq_cr = cr(iq_data, fs_out)
    if plot and plot_save_dir:
        plot_constellation(iq_cr,  sample_skip=1, title="CR Constellation", save=True,
                         save_path=os.path.join(plot_save_dir, 'iq_cr.png'))
        plot_eye_diagram(iq_cr[::2], title="CR Eye Diagram", save=True,
                         save_path=os.path.join(plot_save_dir, 'iq_cr_eye.png'))
        
    iq_rrc = rrc(iq_cr, fs_out,baud_rate)
    if plot and plot_save_dir:
        plot_constellation(iq_rrc,  sample_skip=1, title="RRC Constellation", save=True,
                         save_path=os.path.join(plot_save_dir, 'iq_rrc.png'))
        plot_eye_diagram(iq_rrc[::2], title="RRC Eye Diagram", save=True,
                         save_path=os.path.join(plot_save_dir, 'iq_rrc_eye.png'))
    
    iq_timed,tau,dtau,error = symbol_timing_recovery(iq_rrc)
    if plot and plot_save_dir:
        plot_constellation(iq_timed,  sample_skip=1, title="Timing Recovery Constellation", save=True,
                         save_path=os.path.join(plot_save_dir, 'iq_timed.png'))
        plot_eye_diagram(iq_timed, title="Timing Recovery Eye Diagram", save=True,
                         save_path=os.path.join(plot_save_dir, 'iq_timed_eye.png'))
        fig, axs = plt.subplots(3, 1, figsize=(10, 8))
        axs[0].plot(tau); axs[0].set_title("Timing Offset (tau)")
        axs[1].plot(dtau); axs[1].set_title("Derivative (Dtau)")
        axs[2].plot(error); axs[2].set_title("Timing Error Signal")
        plt.tight_layout()
        plt.savefig(os.path.join(plot_save_dir, 'str_metrics.png'))
        plt.close()
    
    iq_costas, phase,freq = costas(iq_timed)
    if plot and plot_save_dir:
        plot_constellation(iq_costas, sample_skip=1, title="Costas Loop Constellation", save=True,
                         save_path=os.path.join(plot_save_dir, 'iq_costas.png'))
        fig, axs = plt.subplots(2, 1, figsize=(10, 8))
        axs[0].plot(phase); axs[0].set_title('Phase output of PLL'); axs[0].grid()
        axs[1].plot(freq); axs[1].set_title('Frequency of PLL'); axs[1].grid()
        plt.tight_layout()
        plt.savefig(os.path.join(plot_save_dir, 'costas_metrics.png'))
        plt.close()

    bit_decode, deg  = decode(iq_costas)
    if plot and plot_save_dir:
        plt.figure()
        plt.title('Differential Decode Angle')
        plt.xlabel('Symbol Number')
        plt.ylabel('Angle (degrees)')
        plt.plot(deg, 'x')
        plt.grid()
        plt.savefig(os.path.join(plot_save_dir, 'angle.png'))
        plt.close()
    
    offset, reverse_order = find_packet_start(bit_decode)

    hex_packets = bits_to_packets(bit_decode, offset, reverse_order)

    valid_hex_packets = validate_packet(hex_packets, output_file=output_file)

    logger.info("Pipeline finished. Valid packets: %d", len(valid_hex_packets))

    return valid_hex_packets


from typing import List
from .logging_config import get_module_logger

logger = get_module_logger(__name__)


# ===========================
# 1. Basic Functions
# ===========================


def fletcher_checksum(
    hex_data: str
) -> str:
    """
    Calculate the fletcher checksum of the input hex string.

    Args:
        hex_data: parsed ORBCOMM packets hex string, including FCS words.
    
    Returns:
        str: two checksum words as uppercase hex (sum2, sum1).
    """
    sum1 = 0
    sum2 = 0

    for xx in range(0, len(hex_data), 2):
        val = int(hex_data[xx:xx+2], 16)
        sum1 = (sum1 + val) % 256
        sum2 = (sum1 + sum2) % 256

    return f"{sum2:02X}{sum1:02X}"


def single_bit_fix(
    hex_data: str
) -> str | None:
    """
    Attempt single bit error correction on a single packet.

    Args:
        hex_data: parsed ORBCOMM packets hex string, including FCS words.
    
    Returns:
        str | None: if the error correction is successful, return the corrected hex string; otherwise, return none.
    """
    num_hex_data = len(hex_data)
    num_bits = num_hex_data * 4

    binary_packet = f"{int(hex_data, 16):0{num_bits}b}"

    for i in range(len(binary_packet)):
        original_bit = binary_packet[i]
        flip_bit = '1' if original_bit == '0' else '0'

        temp_bits = binary_packet[:i] + flip_bit + binary_packet[i+1:]

        temp_hex = f"{int(temp_bits, 2):0{num_hex_data}X}"

        if fletcher_checksum(temp_hex) == '0000':
            return temp_hex
    return None


# ===========================
# 2. Composite Function
# ===========================




def validate_packet(
    hex_data: List[str],
    output_file: str | None = "orbdemod_packets.txt" ,
) -> List[str]:
    """
    Verify the fletcher checksum of the hex packet, attempt single-bit error correction, and save the result.

    Args:
        hex_data: parsed ORBCOMM packets hex string, including FCS words.
        output_file: the storage path of the valid packets; if it is None, the file will not be saved.

    Returns:
        valid_packets: List of packets that have passed the fletcher checksum or have been corrected.

    """
    logger.info("Packets integrity check (fletcher checksum), 1-bit ECC and save the file.")

    valid_packets = []
    fixed_count = 0
    error_count = 0
    total_packets = len(hex_data)

    for packet in hex_data:
        status = "FAIL"
        final_packet = packet

        if fletcher_checksum(packet) == '0000':
            status = "OK"
        else:
            corrected_packet = single_bit_fix(packet)

            if corrected_packet is not None:
                final_packet = corrected_packet
                status = "FIXED"
                fixed_count +=1
        if status == "FAIL":
            error_count +=1
            logger.warning("Fletcher checksum fails and ECC is unsuccessful or disabled. Dropping packet: %s.", packet)
        else:
            valid_packets.append(final_packet)
            if status == "FIXED":
                logger.info("Packet successfully fix: %s.", final_packet)
    if output_file is not None:
        logger.info("Writing valid and corrected packets to: %s.", output_file)

        with open(output_file, 'w')as f_out:
            for packet in valid_packets:
                f_out.write(packet + '\n')

    valid_count = len(valid_packets)
    ok_count = valid_count - fixed_count

    per = 0.0
    if total_packets > 0:
        per = float(error_count) / total_packets * 100

    
    
    logger.info("_" * 50)
    logger.info("Fletcher checksum and 1-bit ECC processing has completed.")
    logger.info("Total packets processed: %d.", total_packets)
    logger.info("Successfully validated/fixed packets: %d.", ok_count + fixed_count)
    logger.info("Packets dropped: %.2f%% (%d).", per, error_count)
    if output_file is None:
        logger.info("Packet saving is skipped as output_file is set to None.")
    logger.info("-" * 50)

    return valid_packets


import numpy as np
import scipy.signal as sp_signal
from .ddc import frequency_mixing
from .logging_config import get_module_logger

logger = get_module_logger(__name__)

# ===========================
# 1. Basic Functions
# ===========================


def v4_freq_offset_estimator(
    iq_data: np.ndarray,
    fs: float,
    fft_length: int = 9600,
    search_range_hz: float = 2000.0
) -> float:
    """
    4-th power spectrum method to estimate residual frequency offset for input IQ data.
    4-th power removes the phase modulation of SDPSK signals, resulting in a distinct peak at 4 times the frequency offset in the spectrum.

    Args:
        iq_data: input IQ signal.
        fs: Input signal sampling frequency.
        fft_length: FFT length for Welch's method.
        search_range_hz: Search range for the peak in Hz.

    Returns:
        estimated_offset: Estimated residual frequency offset in Hz.
    """
    iq_power_4 = np.power(iq_data, 4)
    f, pxx = sp_signal.welch(iq_power_4, fs=fs, nperseg=fft_length, return_onesided=False)
    f_shifted = np.fft.fftshift(f)
    pxx_shifted = np.fft.fftshift(pxx)

    resolution = fs / fft_length
    search_range_bins = int(search_range_hz / resolution)
    center_bin = len(f_shifted) // 2

    start_bin = center_bin - search_range_bins
    end_bin = center_bin + search_range_bins
    
    if start_bin < 0 or end_bin >= len(pxx_shifted):
        logger.warning("Search range exceeds spectrum boundaries, searching entire spectrum instead.")
        start_bin, end_bin = 0, len(pxx_shifted)

    peak_index_local = np.argmax(pxx_shifted[start_bin:end_bin])
    peak_index_global = start_bin + peak_index_local
    
    if 0 < peak_index_global < len(f_shifted) - 1:
        L = pxx_shifted[peak_index_global - 1]
        C = pxx_shifted[peak_index_global]
        R = pxx_shifted[peak_index_global + 1]

        denominator = (L - 2.0 * C + R)
        if np.abs(denominator) > 1e-12:
            delta = 0.5 * (L - R) / denominator
            freq_at_peak = f_shifted[peak_index_global] + delta * resolution
        else:
            freq_at_peak = f_shifted[peak_index_global]

    else:
        freq_at_peak = f_shifted[peak_index_global]
    
    
    estimated_offset = freq_at_peak / 4.0
    
    logger.info(f"Estimated residual frequency offset: {estimated_offset:.4f} Hz.")

    return estimated_offset


# ===========================
# 2. Composite Function
# ===========================

def carrier_error_recovery(
    iq_data: np.ndarray,
    fs: float,
    num_points_for_fft: int = 50000,
    fft_length: int = 9600,
    search_range_hz: float = 2000.0
) -> np.ndarray:
    """
    ORBCOMM v4 carrier frequency error recovery using 4-th power spectrum method.
    This method is used to eliminate signficant frequency drift.

    Args:
        iq_data: input IQ signal.
        fs: Input signal sampling frequency (Hz).
        num_points_for_fft: Number of points to use for frequency offset estimation.
        fft_length: FFT length for Welch's method.
        search_range_hz: Search range for the peak in Hz.

    Returns:
        iq_data_mixing: IQ signal after frequency mixing to correct the frequency offset.
    """
    logger.info("Starting carrier frequency error recovery process.")

    num_points_for_fft = min(len(iq_data), num_points_for_fft)
    iq_data_fft = iq_data[:num_points_for_fft]
    estimated_offset = v4_freq_offset_estimator(
        iq_data=iq_data_fft,
        fs=fs,
        fft_length=fft_length,
        search_range_hz=search_range_hz
    )
    iq_data_mixing, _ = frequency_mixing(
        data=iq_data,
        freq_lo=estimated_offset,
        fs=fs   
    )

    logger.info("Carrier frequency error recovery has completed.")

    return iq_data_mixing



from .logging_config import enable_logging

from .ddc import (
    digital_down_converter,
    normalize_signal,
    frequency_mixing,
    decimate_iir,
    decimate_fir
)

ddc = digital_down_converter

from .cr import (
    v4_freq_offset_estimator,
    carrier_error_recovery    
)

cr = carrier_error_recovery


from .rrc import (
    rrcosfilter,
    apply_rrc_match_filter
)

rrc = apply_rrc_match_filter

from .timing_recovery import (
    symbol_timing_recovery,
    farrow_interpolator,
    gardner_ted,
    update_buffer
)


from .costas import (
    costas_phase_recovery,
    four_quadrant_detector
)

costas = costas_phase_recovery

from .decode import differential_decode

decode = differential_decode

from .packet_utils import(
    OrbcommPacketType,
    find_packet_start,
    bits_to_packets
)

from .fletcher_ecc_save import (
    fletcher_checksum,
    single_bit_fix,
    validate_packet
)

from .plotting import (
    plot_constellation,
    plot_eye_diagram
)


from .pipeline import orbdemod

__version__ = "0.1.0"

__all__ = [
    "enable_logging",

    "digital_down_converter",
    "ddc",
    "normalize_signal",
    "frequency_mixing",
    "decimate_iir",
    "decimate_fir",

    "carrier_error_recovery",
    "cr",
    "v4_freq_offset_estimator",

    "rrcosfilter",
    "apply_rrc_match_filter",
    "rrc",

    "symbol_timing_recovery",
    "farrow_interpolator",
    "gardner_ted",
    "update_buffer",

    "costas_phase_recovery",
    "four_quadrant_detector",
    "costas",

    "differential_decode",
    "decode",

    "OrbcommPacketType",
    "find_packet_start",
    "bits_to_packets",

    "fletcher_checksum",
    "single_bit_fix",
    "validate_packet",

    "plot_constellation",
    "plot_eye_diagram",

    "orbdemod"
]


import numpy as np
from typing import Tuple
from .logging_config import get_module_logger

logger = get_module_logger(__name__)


# ===========================
# 1. Basic Functions
# ===========================


def four_quadrant_detector(
    sample: np.complex64
) -> float:
    """
    SDPSK Costas loop phase error detector.

    Args:
        sample: the IQ sample rotated by the current phase estimate.
    
    Returns:
        phase_error: the phase error estimate (proportional to phase mismatch).
    """
    y_I = sample.real
    y_Q = sample.imag

    phase_error = np.sign(y_I) * y_Q - np.sign(y_Q) * y_I

    return phase_error



# ===========================
# 2. Composite Function
# ===========================

def costas_phase_recovery(
    iq_data: np.ndarray,
    alpha: float = 0.03,
    beta_ratio: float = 0.2,
    beta: float = None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Perform carrier phase recovery using a second-order Costas loop for SDPSK signals.

    Args:
        iq_data: input IQ signal after timing recovery.
        alpha: loop proportional gain (phase correction).
        beta_ratio: ratio to calculating integral gain (frequency correction).
        beta: Integral gain (optional).
        
    Returns:
        costas_iq: IQ signal after carrier phase correction.
        phase_est: history of the phase estimate at each input sample.
        freq_out_hist: history of the frequency estimate at each input sample.
    """
    if beta is None:
        beta = beta_ratio * alpha**2

    logger.info(
        "Starting Costas Loop Phase Recovery.Alpha: %f, Beta: %f", 
        alpha, beta
    )
    

    n = len(iq_data)
    phase_est = np.zeros(n + 1)
    frequency_out = 0.0
    freq_out_hist = np.zeros(n)

   

    for i, sample in enumerate(iq_data):
   
        corrected_sample = sample * np.exp(-1j * phase_est[i]) 
        phase_error = four_quadrant_detector(corrected_sample)
              
        frequency_out += beta * phase_error
        phase_est[i+1] = phase_est[i] + alpha * phase_error + frequency_out

        freq_out_hist[i] = frequency_out
        
    costas_iq = iq_data * np.conj(np.exp(1j*phase_est[:-1]))
    
    logger.info("Costas Loop Phase Recovery has completed. Final frequency offset estimate: %.4f rad/symbol.", frequency_out)
    
    return costas_iq, phase_est,freq_out_hist


import numpy as np
import scipy.signal as sp_signal
from .logging_config import get_module_logger

logger = get_module_logger(__name__)

# From: https://github.com/veeresht/CommPy/blob/master/commpy/filters.py （same in https://github.com/fbieberly/ORBCOMM-receiver/blob/master/helpers.py）
def rrcosfilter(N, alpha, Ts, Fs):
    """
    Generates a root raised cosine (RRC) filter (FIR) impulse response.

    Parameters
    ----------
    N : int
        Length of the filter in samples.

    alpha: float
        Roll off factor (Valid values are [0, 1]).

    Ts : float
        Symbol period in seconds.

    Fs : float
        Sampling Rate in Hz.

    Returns
    ---------
    h_rrc : 1-D ndarray of floats
        Impulse response of the root raised cosine filter.

    time_idx : 1-D ndarray of floats
        Array containing the time indices, in seconds, for
        the impulse response.
    """
    N = int(N)
    T_delta = 1/float(Fs)
    time_idx = ((np.arange(N)-N/2))*T_delta
    sample_num = np.arange(N)
    h_rrc = np.zeros(N, dtype=float)

    for x in sample_num:
        t = (x-N/2)*T_delta
        if t == 0.0:
            h_rrc[x] = 1.0 - alpha + (4*alpha/np.pi)
        elif alpha != 0 and t == Ts/(4*alpha):
            h_rrc[x] = (alpha/np.sqrt(2))*(((1+2/np.pi)* \
                    (np.sin(np.pi/(4*alpha)))) + ((1-2/np.pi)*(np.cos(np.pi/(4*alpha)))))
        elif alpha != 0 and t == -Ts/(4*alpha):
            h_rrc[x] = (alpha/np.sqrt(2))*(((1+2/np.pi)* \
                    (np.sin(np.pi/(4*alpha)))) + ((1-2/np.pi)*(np.cos(np.pi/(4*alpha)))))
        else:
            h_rrc[x] = (np.sin(np.pi*t*(1-alpha)/Ts) +  \
                    4*alpha*(t/Ts)*np.cos(np.pi*t*(1+alpha)/Ts))/ \
                    (np.pi*t*(1-(4*alpha*t/Ts)*(4*alpha*t/Ts))/Ts)

    return time_idx, h_rrc



def apply_rrc_match_filter(
    iq_data: np.ndarray,
    fs: float,
    baud_rate: float,
    span_symbols: int = 8
) -> np.ndarray:
    """
    ORBCOMM RRC matched filtering.

    Args:
        iq_data: input IQ signal.
        fs: Input signal sampling frequency (Hz).
        baud_rate: Symbol rate (baud).
        span_symbols: Filter half-span in symbols.

    Returns:
        matched_filtered_signal: IQ signal after RRC matched filtering.
    """
    alpha = 0.4
    sps = fs / baud_rate
    rrc_num_taps = int(sps * span_symbols * 2 + 1)

    logger.info(
        "Starting RRC matched filtering. Baud: %.2f, SPS: %.2f, Taps: %d (Span: %d symbols, Alpha: %.2f).",
        baud_rate, sps, rrc_num_taps, span_symbols, alpha
    )

    rrc_taps = rrcosfilter(
        rrc_num_taps,
        alpha,
        Ts=1.0,
        Fs=sps
    )[1]

    matched_filtered_signal = sp_signal.lfilter(rrc_taps, 1.0, iq_data)

    logger.info("RRC matched filtering has completed.")

    return matched_filtered_signal


import numpy as np
from typing import Tuple
from .logging_config import get_module_logger

logger = get_module_logger(__name__)

def differential_decode(
    iq_phase_locked: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Perform differential decoding for SDPSK signals.
    
    Args:
        iq_phase_locked: IQ samples after Costas loop phase recovery.

    Returns:
        decode_bits: final decoded bit sequence.
        phase_diffs_deg: history of phase differences in degrees.
    """
    logger.info(
        "Starting differential decoding."
    )
    iq_phase_locked /= np.median(np.abs(iq_phase_locked))

    symbols = iq_phase_locked

    phase_diffs_rad = np.angle(symbols[1:] * np.conj(symbols[:-1]))

    decoded_bits = (phase_diffs_rad > 0).astype(np.uint8)

    phase_diffs_deg = np.rad2deg(phase_diffs_rad)

    avg_abs_phase = np.mean(np.abs(phase_diffs_deg))
    logger.info(
        "Differential decoding has completed. Bits generated: %d. Mean absolute phase diff: %.4f deg.",
        len(decoded_bits), avg_abs_phase
    )

    return decoded_bits, phase_diffs_deg


import numpy as np
from typing import Tuple, List
from .logging_config import get_module_logger

logger = get_module_logger(__name__)

# from https://github.com/fbieberly/ORBCOMM-receiver/blob/master/orbcomm_packet.py

class OrbcommPacketType:
    DEFAULT_PACKET_SIZE_BITS = 12 * 8

    packet_dict = {
        'Sync':{
            'header':'01100101',
            'hex_header':'65',
            'message_parts':[
                ('code',(0,6)),
                ('sat_id',(6,8)),
            ],
        },

        'Message':{
            'header':'00011010',
            'hex_header':'1A',
            'message_parts':[
                ('msg_total_length', (2, 3)), 
                ('msg_packet_num', (3, 4)),
                ('data', (4, 20)),
            ],
        },

        'Uplink_info':{
            'header':'00011011',
            'hex_header':'1B',
            'message_parts':[
                ('msg_total_length', (2, 3)),
                ('msg_packet_num', (3, 4)),
                ('data', (4, 20)),
            ],
        },

        'Downlink_info':{
            'header':'00011100',
            'hex_header':'1C',
            'message_parts':[
                ('msg_total_length', (2, 3)),
                ('msg_packet_num', (3, 4)),
                ('data', (4, 20)),
            ],
        },

        'Network':{
            'header':'00011101',
            'hex_header':'1D',
            'message_parts':[
                ('msg_total_length', (2, 3)),
                ('msg_packet_num', (3, 4)),
                ('data', (4, 20)),
            ],

        },

        'Fill':{
            'header':'00011110',
            'hex_header':'1E',
            'message_parts':[ 
                ('data', (2, 20)),
            ],

        },

        'Ephemeris':{
            'header':'00011111',
            'hex_header':'1F',
            'message_parts':[
                ('sat_id', (2, 4)),  
                ('data', (4, 46)),  
            ],

        },

        'Orbital':{
            'header':'00100010',
            'hex_header':'22',
            'message_parts':[
            ],

        },
    }


    @staticmethod
    def get_packet_headers() -> List[str]:
        """
        Extract binary header strings for all known data packets.
        """
        return [data['header'] for data in OrbcommPacketType.packet_dict.values()]


def find_packet_start(
    decode_bits: np.ndarray
) -> Tuple[int, bool]:
    """
    The best packet starting point and byte order are determined by calculating the number of matches with the known headers on all possible offset.

    Args:
        decode_bits: the demodulated bit data (not string yet).

    Return:
        best_offset: the optimal starting bit offset.
        use_reversed_bits: True means bits are reversed (LSB-first).
    """
    bit_string = ''.join(map(str, decode_bits))
    logger.info("Starting search for packets best offset and bits order.")

    packet_headers = OrbcommPacketType.get_packet_headers()
    packet_size_bits = OrbcommPacketType.DEFAULT_PACKET_SIZE_BITS


    normal_scores = np.zeros(packet_size_bits)
    rev_scores = np.zeros(packet_size_bits)

    for offset in range(packet_size_bits):
        for i in range(offset, len(bit_string) - 8, packet_size_bits):
            header_candidate = bit_string[i : i+8]
            
            if header_candidate[::-1] in packet_headers:
                rev_scores[offset] += 1
            if header_candidate in packet_headers:
                normal_scores[offset] += 1
    
    use_reversed_bits = np.max(rev_scores) > np.max(normal_scores)

    if use_reversed_bits:
        best_offset = np.argmax(rev_scores)
        logger.info(
            "Optimal offset found at: %d (Bits are REVERSED).", 
            best_offset
        )
    else:
        best_offset = np.argmax(normal_scores)
        logger.info(
            "Optimal offset found at: %d (Bits are NOT reversed).", 
            best_offset
        )

    return best_offset, use_reversed_bits



def bits_to_packets(
    decode_bits: np.ndarray,
    offset: int,
    reverse_order: bool
) -> List[str]:
    """
    Parse Orbcomm packets from the bit stream based on the best offset and bits order.

    Args:
        decode_bits: the demodulated bit data (not string yet).
        offset: the optimal starting bit offset.
        reverse_order: True means bits are reversed (LSB-first).

    Returns:
        List[str]: the list of parsed packets, each items is a hex string.
    """
    bit_string = ''.join(map(str, decode_bits))
    logger.info("Parsing bits stream into hex packets.")

    packets = []

    protocol_dict = OrbcommPacketType.packet_dict
    packet_size_bits = OrbcommPacketType.DEFAULT_PACKET_SIZE_BITS


    eph_header = protocol_dict['Ephemeris']['hex_header']
    i = offset

    while i <= len(bit_string) - packet_size_bits:
        header_bits = bit_string[i : i+8]
        if reverse_order:
            header_bits = header_bits[::-1]


        header_hex = f"{int(header_bits, 2):02X}"
        current_packet_len_bits = packet_size_bits
        
        if header_hex == eph_header:
            current_packet_len_bits = 24 * 8
            logger.debug("Detect ephemeris packet (Type 1F).")
        
        if i + current_packet_len_bits > len(bit_string):
            break
            
        packet_hex = ""
        for j in range(0, current_packet_len_bits, 8):
            byte_bits = bit_string[i+j : i+j+8]
            if reverse_order:
                byte_bits = byte_bits[::-1]
            packet_hex += f"{int(byte_bits, 2):02X}"
        
        packets.append(packet_hex)
        
        i += current_packet_len_bits
        
    logger.info("Successfully parse %d packets from the bits stream.", len(packets))
    
    return packets


import logging

ORBDEMOD_ROOT_NAME = 'orbdemod'
ORBDEMOD_ROOT_LOGGER = logging.getLogger(ORBDEMOD_ROOT_NAME)

ORBDEMOD_ROOT_LOGGER.addHandler(logging.NullHandler())


def get_module_logger(name: str) -> logging.Logger:
    if not name.startswith(ORBDEMOD_ROOT_NAME):
        name = f"{ORBDEMOD_ROOT_NAME}.{name}"
    return logging.getLogger(name)

def enable_logging(level=logging.INFO):

    logger = logging.getLogger(ORBDEMOD_ROOT_NAME)
    if not any(isinstance(h, logging.StreamHandler) for h in logger.handlers):
        ch = logging.StreamHandler()
        formatter = logging.Formatter('[%(name)s: %(levelname)s] %(message)s')
        ch.setFormatter(formatter)

        logger.addHandler(ch)

    logger.setLevel(level)



import numpy as np
import scipy.signal as sp_signal
from typing import Tuple
from .logging_config import get_module_logger

logger = get_module_logger(__name__)

# ===========================
# 1. Basic Functions
# ===========================

def normalize_signal(
    data: np.ndarray
) -> np.ndarray:
    """
    Normalize the signal.
    """
    data = data.astype(np.float32)
    data_normalized = data / np.median(np.abs(data))

    return data_normalized


def frequency_mixing(
    data: np.ndarray,
    freq_lo: float,
    fs: float,
    initial_phase: float = 0.0
) -> Tuple[np.ndarray, float]:
    """
    Mix the input voltage signal with a local oscillator to shift its frequency.

    Args:
        data: input voltage signal.
        freq_lo: local oscillator frequency (the frequency to shift by).
        fs: sampling frequency of the input signal.
        initial_phase: initial phase of the local oscillator in radians.

    Returns:
        mixed_signal: frequency-shifted IQ signal.
        final_phase: final phase of the local oscillator after processing the input signal, which can be used for next segment processing.
    """
    num_samples = len(data)
    
    dphi = 2 * np.pi * freq_lo / fs
    phase = initial_phase + dphi * np.arange(num_samples)
    lo_signal = np.exp(-1j * phase)
    mixed_signal = data * lo_signal

    final_phase = initial_phase + dphi * num_samples
    final_phase = np.angle(np.exp(1j * final_phase))

    return mixed_signal, final_phase


def decimate_iir(
    data: np.ndarray,
    fs: float,
    fs_target: float,
    cutoff_ratio: float = 2.2,
    order: int = 6
) -> np.ndarray:
    """
    Use an IIR Butterworth SOS filter for anti-aliasing and decimation.
    Fast and efficient, with a nonlinear phase; zero phase can be obtained using sosfiltfilt.
    Intended for first-stage significant downsampling.
    Notice: This method is suitable for offline only. For real-time processing, use sosfilt instead of sosfiltfilt.

    Args:
        data: input IQ signal.
        fs: original sampling frequency of the input signal.
        fs_target: target sampling frequency after decimation.
        cutoff_ratio: ratio to determine the cutoff frequency for the low-pass filter.
        order: order of the Butterworth filter.

    Returns:
        filtered_data: decimated IQ signal.
    """
    decimation_factor = int(fs / fs_target)
    cutoff_freq = fs_target / cutoff_ratio
    sos = sp_signal.butter(order, cutoff_freq, btype='low', fs=fs, output='sos')
    filtered_data = sp_signal.sosfiltfilt(sos, data)

    return filtered_data[::decimation_factor]


def decimate_fir(
    data: np.ndarray,
    fs: float,
    fs_target: float,
    cutoff_ratio: float = 2.4,
    num_taps: int = 201
) -> np.ndarray:
    """
    Use an FIR filter for anti-aliasing and decimation.
    Linear phase, large amount of computation.
    Notice: This method is suitable for offline only. For real-time processing, use lfilter instead of filtfilt.

    Args:
        data: input IQ signal.
        fs: original sampling frequency of the input signal.
        fs_target: target sampling frequency after decimation.
        cutoff_ratio: ratio to determine the cutoff frequency for the low-pass filter.
        num_taps: number of taps (coefficients) in the FIR filter.

    Returns:
        filtered_data: decimated IQ signal.
    """
    decimation_factor = int(fs / fs_target)
    cutoff_freq = fs_target / cutoff_ratio

    fir_filter = sp_signal.firwin(num_taps, cutoff = cutoff_freq, fs=fs)

    filtered_data = sp_signal.filtfilt(fir_filter, 1.0, data)

    return filtered_data[::decimation_factor]


# ===========================
# 2. Composite Function
# ===========================


def digital_down_converter(
    data: np.ndarray,
    freq_lo: float,
    fs_in: float,
    fs_mid: float,
    fs_out: float,
    initial_phase: float = 0.0
) -> Tuple[np.ndarray, float]:
    """
    Perform digital down conversion (DDC) pipeline on the input voltage signal from ORBCOMM satellites.

    Args:
        data: input voltage signal.
        freq_lo: local oscillator frequency (the frequency to shift by).
        fs_in: original sampling frequency of the input signal.
        fs_mid: intermediate sampling frequency after first decimation.
        fs_out: final target sampling frequency after second decimation.
        initial_phase: initial phase of the local oscillator in radians.

    Returns:
        data_final: final down-converted IQ signal.
        next_phase: final phase of the local oscillator after processing the input signal, which can be used for next segment processing.
    """
    logger.info(
        "Starting Digital Down Converter (DDC) pipeline. Fs: %.2f Hz -> %.2f Hz -> %.2f Hz.",
        fs_in, fs_mid, fs_out
    )
    logger.info("Local Oscillator (LO) frequency: %.2f Hz. Initial Phase: %.2f rad.", freq_lo, initial_phase)

    norm_data = normalize_signal(data)

    mixed_data, next_phase = frequency_mixing(norm_data, freq_lo, fs_in, initial_phase)

    dc_offset = np.mean(mixed_data)

    logger.info("DC Offset Removal.")
    mixed_data_center = mixed_data - dc_offset

    logger.info("DDC Step 1/2: Applying IIR Decimation (Fs: %.2f Hz -> %.2f Hz).", fs_in, fs_mid)
    data_mid = decimate_iir(data=mixed_data_center, fs=fs_in, fs_target=fs_mid)

    logger.info("DDC Step 2/2: Applying FIR Decimation (Fs: %.2f Hz -> %.2f Hz).", fs_mid, fs_out)
    data_final = decimate_fir(data=data_mid, fs=fs_mid, fs_target=fs_out)

    logger.info("DDC pipeline has completed. Final Phase: %.2f rad.", next_phase)

    return data_final, next_phase








import numpy as np
import matplotlib.pyplot as plt
from typing import Tuple
from .logging_config import get_module_logger

logger = get_module_logger(__name__)

def plot_constellation(
    iq_signal: np.ndarray,
    sample_skip: int = 100,
    title: str = 'Constellation Diagram',
    show_plot: bool = False,
    save_path: str = None,
    save: bool = False
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Draw the constellation diagram of IQ signal.

    Args:
        iq_signal: IQ signal.
        sample_skip: sampling interval.
        title: chart title.
        show_plot: whether to immediately call plt.show() to display the chart (not display by default).
        save_path: the path to save the chart.
        save: whether to immediately call plt.savefig() to save the chart (not save by default).

    Returns:
        fig: the figure object of matplotlib.
        ax: the axes object of matplotlib.

    """
    plot_data = iq_signal[::sample_skip]

    fig, ax = plt.subplots(figsize=(8, 8))

    ax.scatter(plot_data.real, plot_data.imag, s=10, alpha=0.6, marker='.')

    ax.set_title(title)
    ax.set_xlabel('I Component (In-phase)')
    ax.set_ylabel('Q Component (Quadrature)')
    ax.grid(True, linestyle='--', alpha=0.5)

    ax.axhline(0, color='red', linewidth=0.8)
    ax.axvline(0, color='red', linewidth=0.8)

    ax.set_aspect('equal', adjustable='box') 
    


    if save:
        if save_path is None:
            logger.warning("save=True but save_path=None, skipping save.")
        else:
            plt.savefig(save_path)

    if show_plot:
        plt.show()

    logger.info("Constellation plot generated with %d data points (skip=%d).", len(plot_data), sample_skip)
    
    return fig, ax
    
    
def plot_eye_diagram(
    iq_data: np.ndarray,
    sample_offset: int = 400,
    segments: int = 10,
    segment_length: int = 64,
    title: str = 'Eye Diagram (Q Component)',
    show_plot: bool = False,
    save_path: str = None,
    save: bool = False
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Draw the eye diagram of IQ signal.

    Args:
        iq_data: IQ signal.
        sample_offset: starting index for eye diagram.
        segments: number of segments to overlay.
        segment_length: number of samples per eye segment.
        title: chart title.
        show_plot: whether to immediately call plt.show() to display the chart (not display by default).
        save_path: the path to save the chart.
        save: whether to immediately call plt.savefig() to save the chart (not save by default).

    Returns:
        fig: the figure object of matplotlib.
        ax: the axes object of matplotlib.
    """
    fig, ax = plt.subplots(figsize=(10, 6))

    current_index = sample_offset
    num_plotted = 0

    while current_index + segment_length <= len(iq_data) and num_plotted < segments:

        segment = iq_data[current_index:current_index + segment_length].imag
        ax.plot(segment, alpha=0.7)

        current_index += segment_length
        num_plotted += 1

    ax.set_title(title)
    ax.set_xlabel('Sample Index')
    ax.set_ylabel('Amplitude (Q Component)')
    ax.grid(True, linestyle='--', alpha=0.5)

    if save:
        if save_path is None:
            logger.warning("save=True but save_path=None, skipping save.")
        else:
            plt.savefig(save_path)

    if show_plot:
        plt.show()

    logger.info("Eye Diagram generated with %d segments (segment_length=%d).", num_plotted, segment_length)

    return fig, ax

